# ***************************************************************************
# Copyright 2014-2020 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import nsx_logging
import os
try:
    import simplejson
except ImportError:
    import json as simplejson

from cli.commands.cli_command_provider import CliCommandProvider
from cli.commands.common.arguments import IntegerArgument
from cli.utils.location_utils import FilePaths
from cli.utils.logging_utils import get_commands_history
from cli.utils.node_utils import is_appliance, is_windows, get_user_id
from cli.utils.rest_utils import local_node_request
from compat23 import compat23_httplib


lg = nsx_logging.getLogger("cli.commands.common.cli_config")

_MAX_CLI_HISTORY_SIZE_VALUE = 2 ** 31 - 1


class GetCliOutputDateTime(CliCommandProvider):
    def __call__(self, cmd_list):
        cli_output_datetime = None
        if is_appliance():
            status, obj, err, _ = local_node_request("GET", "/node")
            if status == compat23_httplib.OK:
                cli_output_datetime = obj.get("cli_output_datetime")
            else:
                lg.error("Request to get 'cli_output_datetime' status failed")
        else:
            fpath = FilePaths.CLI_CONFIG_FILE if is_windows()\
                        else FilePaths.HOST_CLI_CONFIG_FILE
            try:
                with open(fpath) as fh:
                    data = simplejson.load(fh)
                cli_output_datetime = data.get("cli_output_datetime")
            except Exception:
                lg.error("Unable to read NSX CLI config file")
        if cli_output_datetime is not None:
            return "OK\n\n%s" % simplejson.dumps(
                    {"cli_output_datetime": cli_output_datetime})
        return "MSG_CLI_CONFIG_GET_OUTPUT_DATETIME_FAILED"


class SetCliOutputDateTime(CliCommandProvider):
    def __call__(self, cmd_list):
        status = True if cmd_list[-1] == "enable" else False
        req_obj = {"cli_output_datetime": status}
        result = False
        if is_appliance():
            status, obj, err, _ = local_node_request("PUT", "/node", req_obj)
            if status == compat23_httplib.OK:
                result = True
            else:
                lg.error("Request to set 'cli_output_datetime' status failed")
        else:
            fpath = FilePaths.CLI_CONFIG_FILE if is_windows()\
                        else FilePaths.HOST_CLI_CONFIG_FILE
            try:
                with open(fpath) as fh:
                    data = simplejson.load(fh)
                data.update(req_obj)
                with open(fpath, "w") as fh:
                    simplejson.dump(data, fh, indent=4)
                result = True
            except Exception:
                lg.exception("Unable to perform IO on NSX CLI config file")
        if result:
            return "OK\n\n%s" % simplejson.dumps("")
        return "MSG_CLI_CONFIG_SET_OUTPUT_DATETIME_FAILED"


# ----------------------------------------------------------------------
# commands history
# ----------------------------------------------------------------------
def _get_command_history_size():
    """Get command history size
    Returns:
        commands history size
    """
    cli_hist_size = None
    fpath = FilePaths.CLI_CONFIG_FILE if (is_windows() or is_appliance())\
                else FilePaths.HOST_CLI_CONFIG_FILE
    try:
        with open(fpath) as fh:
            data = simplejson.load(fh)
        cli_hist_size = data.get("cli_history_size")
    except Exception:
        lg.error("Unable to read NSX CLI config file")
    return cli_hist_size


class CliHistorySizeArgument(IntegerArgument):
    """Verify argument to validate cli history size value."""
    _MIN_VALUE = 0
    _MAX_VALUE = _MAX_CLI_HISTORY_SIZE_VALUE


class GetCommandHistory(CliCommandProvider):
    def __call__(self, cmd_list):
        cli_hist_size = _get_command_history_size()
        if cli_hist_size is None:
            return "MSG_NODE_GET_COMMAND_HISTORY_FAILED"
        else:
            cli_hist_size = (cli_hist_size + 1) if cli_hist_size \
                                                else cli_hist_size
            userid = get_user_id()
            result = get_commands_history(userid, cli_hist_size)
            result = result[:-1]
            # add sequence number for each command
            for seq_number, cmd in enumerate(result, start=1):
                cmd["seq_number"] = seq_number
            return "OK\n\n%s" % simplejson.dumps(result)


class SetCommandHistoryLimit(CliCommandProvider):
    def __call__(self, cmd_list):
        req_obj = {"cli_history_size": int(cmd_list[3])}
        result = False
        if is_appliance():
            status, obj, err, _ = local_node_request("PUT", "/node", req_obj)
            if status == compat23_httplib.OK:
                result = True
            else:
                lg.error("Request to set 'cli_hist_size' status failed")
        else:
            fpath = FilePaths.CLI_CONFIG_FILE if is_windows()\
                        else FilePaths.HOST_CLI_CONFIG_FILE
            try:
                with open(fpath) as fh:
                    data = simplejson.load(fh)
                data.update(req_obj)
                with open(fpath, "w") as fh:
                    simplejson.dump(data, fh, indent=4)
                result = True
            except Exception:
                lg.exception("Unable to perform IO on NSX CLI config file")
        if result:
            return "OK\n\n%s" % simplejson.dumps("")
        return "MSG_NODE_SET_COMMAND_HISTORY_FAILED"
