# **********************************************************************
# Copyright 2017 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
import logging
from logging.config import dictConfig
import nsx_logging
from nsx_logging.handlers import GetRotatingFileHandler


class DictConfigurator(logging.config.DictConfigurator):
    """
    This class is used to override configurations for nsx-specific
    handlers.
    """

    def configure_handler(self, config):
        """
        This method is used to generate a log handler dynamically based on config.
        """

        # validate some 'must have' parameters
        if "filename" not in config or "class" in config:
            return super(DictConfigurator, self).configure_handler(config)

        # get handler arguments
        formatter = config.pop("formatter", None)
        if formatter:
            try:
                formatter = self.config["formatters"][formatter]
            except Exception as e:
                raise ValueError("Unable to set formatter "
                                 "%r: %s" % (formatter, e))
        level = config.pop("level", None)
        filters = config.pop("filters", None)
        kwargs = dict([(k, config[k]) for k in config])

        # create the handler object
        result = GetRotatingFileHandler(**kwargs)
        if formatter:
            result.setFormatter(formatter)
        if level is not None:
            result.setLevel(logging._checkLevel(level))
        if filters:
            self.add_filters(result, filters)

        return result


# override configurator defined in logging.config
logging.config.dictConfigClass = DictConfigurator


def _loadyaml(filename):
    """
    Read the config file in yaml format and return a config dictionary.
    If error happens, return None.
    """
    try:
        import yaml
    except ImportError:
        nsx_logging.getLogger().error("python-yaml is not installed")
        return None

    try:
        with open(filename, "r") as stream:
            return yaml.load(stream)
    except:
        try:
            nsx_logging.getLogger().error(
                "python logging configure file '{}' error".format(filename))
        except:
            pass
        return None


def fileConfig(filename):
    """
    Load a logging config file in yaml format.
    Use logging.config.dictConfig to load the configurations.
    """

    # support for common config formats
    supported_ext = {
        "yaml": _loadyaml,
        "yml": _loadyaml,
    }

    if not isinstance(filename, str):
        return
    ext = filename.split(".")[-1]
    if ext not in supported_ext:
        try:
            nsx_logging.getLogger().error(
                "python logging config format '{}' not supported".format(ext))
        except:
            pass
        return

    # parse config file
    confDict = supported_ext[ext](filename)
    if confDict is None:
        return

    confDict["formatters"]["rfc5424"]["comp"] = nsx_logging._get_nsx_node_type()

    dictConfig(confDict)
    return

