# ***************************************************************************
# Copyright 2017-2020 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
import getpass
import logging
import os
import socket
import sys
import time


class StructuredDataFormatter(logging.Formatter):
    converter = time.gmtime  # Overload logging.Formatter.converter

    def __init__(self, msgid="-", comp="nsx-manager", subcomp="python",
                 node_type="nsx-manager"):
        self.msgid = msgid
        self.msgid_list = []
        self.sd_dict = {}
        self.comp = comp
        self.subcomp = subcomp
        self.username = getpass.getuser()
        self.node_type = node_type
        # Get supported message id and structured data for validation
        nsx_common_path = os.path.dirname(os.path.dirname(os.path.dirname(
            os.path.abspath(__file__))))
        msgid_txt = os.path.join(nsx_common_path, "logging", "message-ids.txt")
        sd_txt = os.path.join(nsx_common_path, "logging", "structured-data.txt")
        try:
            with open(msgid_txt) as msgids:
                self.msgid_list = msgids.read().splitlines()
            with open(sd_txt) as sd:
                for line in sd:
                    tokens = list(filter(None, line.strip().split(":")))
                    key = tokens[0]
                    values = tokens[1].split(",") if len(tokens) == 2 else []
                    self.sd_dict[key] = values
        except:
            # if message-ids.txt or structured-data.txt is missing, leave the
            # msgid_list and sd_dict empty
            pass
        if self.node_type == "nsx-esx":
            fmt = self._esx_structured_data_format()
        else:
            fmt = self._structured_data_format()
        datefmt = "%Y-%m-%dT%H:%M:%S"
        super(StructuredDataFormatter, self).__init__(fmt, datefmt)

    def format(self, record):
        sd_tags, sd_map = "", record.__dict__.get("tags", {})
        if isinstance(sd_map, dict):
            invalid = False
            # 1. enter this block and fill "msgid", "comp" and "subcomp", which
            # are three fields needed for every log record
            msgid = sd_map.pop("msgid", self.msgid)
            # If invalid message id is provided, use '-' instead
            if self.msgid_list and msgid not in self.msgid_list:
                msgid = "-"
                invalid = True
            comp = sd_map.pop("comp", self.comp)
            # If invalid comp value is provided, use node type instead
            if self.sd_dict and comp not in self.sd_dict["comp"]:
                comp = self.node_type
                invalid = True
            record.__dict__["msgid"] = msgid
            record.__dict__["comp"] = comp
            record.__dict__["subcomp"] = sd_map.pop("subcomp", self.subcomp)
            record.__dict__["username"] = sd_map.pop("username", self.username)
            sd_items = ['%s="%s"' % (k, v) for k, v in list(sd_map.items())
                if not self.sd_dict or (k in self.sd_dict and
                (not self.sd_dict[k] or v in self.sd_dict[k]))]
            if invalid or len(sd_items) != len(sd_map.keys()) and \
                    "invalid" not in self.sd_dict:
                sd_items.append('invalid="true"')
            if len(sd_items) > 0:
                sd_tags = " %s" % " ".join(sd_items)
        elif isinstance(sd_map, str):
            # 2. second time enter method 'format'
            # now "msgid", "comp" and "subcomp" are already filled
            sd_tags = sd_map
        record.__dict__["tags"] = sd_tags
        return super(StructuredDataFormatter, self).format(record)

    def _structured_data_format(self):
        """Construct a structured data styled format string.
        Message format that complies with RFC5424.
        """
        log_format = (
            "1 %(asctime)s.%(msecs)03dZ {} NSX %(process)d %(msgid)s "
            "[nsx@6876 comp=\"%(comp)s\" subcomp=\"%(subcomp)s\" "
            "username=\"%(username)s\" level=\"%(levelname)s\"%(tags)s] "
            "%(message)s")
        hostname = socket.gethostname()
        return log_format.format(hostname)

    def _esx_structured_data_format(self):
        """Construct a structured data styled format string.
        Message format that is similar to RFC5424 but with header generated
        by vmsyslogd.
        """
        # Exclude timestamp and hostname on ESX because vmsyslogd will prepend
        # a default header.
        log_format = (
            "NSX %(process)d %(msgid)s [nsx@6876 comp=\"%(comp)s\" "
            "subcomp=\"%(subcomp)s\" username=\"%(username)s\" "
            "level=\"%(levelname)s\"%(tags)s] %(message)s")
        return log_format


class UtcFormatter(logging.Formatter):
    converter = time.gmtime
