# Copyright (C) 2019 VMware, Inc. All rights reserved.
# -- VMware Confidential
import time
import sha.config
from vmware.nsx.sha import sha_pb2

_logger = sha.config.get_logger(__name__)


def should_report(pre_report_msg, pre_report_time, pre_reported_by_change, current_msg,
                  report_interval, smallest_report_interval_if_change):
    """
        Decide whether should report to MP. If yes, also return if this report is triggered by msg change.

        There are 3 cases should report.
        1. current report is the 1st report
        2. time interval between two reports is larger than report_interval
        3. current report msg is different with pre-report, and time interval between two
        reports by change is larger than smallest_report_interval_if_change.
    """
    # report if this is the first report
    if not pre_report_msg:
        _logger.debug("Do report as this is the first report msg")
        return True, False

    # report by interval - time interval between two reports is larger than report_interval
    now = time.monotonic()
    if now - pre_report_time >= report_interval:
        _logger.debug("Do report as time interval between two reports is larger than report_interval:{}s"
                      .format(report_interval))
        return True, False

    # report by change -
    # report if current msg is different with previous msg report by report_interval
    # OR interval between two reports by change is larger than smallest_report_interval_if_change
    if current_msg != pre_report_msg:
        if not pre_reported_by_change:
            _logger.debug("Do report as current msg is different with previous msg reported by report_interval:{}s"
                          .format(report_interval))
            return True, True
        if now - pre_report_time >= smallest_report_interval_if_change:
            _logger.debug("Do report as interval between two reports by change is larger than "
                          "smallest_report_interval_if_change:{}s".format(smallest_report_interval_if_change))
            return True, True
        _logger.debug("Not report as status flapping in {}s, skipping msg:{}"
                      .format(smallest_report_interval_if_change, sha_pb2.ShaReportMsg().FromString(current_msg)))

    _logger.debug("Not report, skipping msg:{}".format(sha_pb2.ShaReportMsg().FromString(current_msg)))
    return False, pre_reported_by_change
