# Copyright (C) 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import json
import fcntl
import sys
import traceback

from sha.contrib.metric.edge_base import EdgeBase
from sha.contrib.metric.utils import timeout_cache

_EDGE_NIC_THROUGHPUT_FILE = "/var/run/vmware/edge/nic_throughput.json"

class EdgeDpStats(EdgeBase):

    def get_dp_cpu_stats(self):
        dp_cpu_stats = []
        dp_cpu_stats_dict = {}


        # Get DP CPU stats
        cmd_get_dp = ['edge-appctl', '-t', '/var/run/vmware/edge/dpd.ctl', 'config/get']
        output_dp = self.run_command_with_time_check(cmd_get_dp)
        dp_stats = json.loads(output_dp)
        sp_core_set = dp_stats["service_corelist"]

        cmd_get_cpu = ['edge-appctl', '-t', '/var/run/vmware/edge/dpd.ctl', 'cpuusage/cycles']
        output_cpu = self.run_command_with_time_check(cmd_get_cpu)
        dp_cpu_stats = json.loads(output_cpu)

        for stat in dp_cpu_stats:
            if str(stat["core"]) not in sp_core_set:
                dp_cpu_stats_dict[stat["core"]] = stat

        self._logger.debug("edge_dp_cpu stats({}): {}".format(sys.getsizeof(dp_cpu_stats_dict), dp_cpu_stats_dict))
        return dp_cpu_stats_dict

    def get_dp_mem_pool_stats(self):
        dp_mem_stats = []
        dp_heap_stats = []
        dp_mem_stats_dict = {}

	# Get DP Mem Pool stats
        cmd_get_stats = ['edge-appctl', '-t', '/var/run/vmware/edge/dpd.ctl', 'mempool/show']
        output = self.run_command_with_time_check(cmd_get_stats)
        dp_mem_stats = json.loads(output)

        for stat in dp_mem_stats:
            dp_mem_stats_dict[stat["name"]] = stat

        heap_cmd = ['edge-appctl', '-t', '/var/run/vmware/edge/dpd.ctl', 'memory/show', 'malloc_heap']
        heap_output = self.run_command_with_time_check(heap_cmd)
        dp_heap_stats = json.loads(heap_output)

        for heap_stat in dp_heap_stats:
            if heap_stat["Heap name"]:
                heap_stat["available_entries"] = heap_stat["Free_size"]
                heap_stat["size"] = heap_stat["Heap_size"]
                dp_mem_stats_dict[heap_stat["Heap name"]] = heap_stat

        self._logger.debug("edge_dp_mem_pool dict stats({}): {}".format(sys.getsizeof(dp_mem_stats_dict), dp_mem_stats_dict))
        return dp_mem_stats_dict


    def get_dp_kni_intf_stats(self):
        dp_kni_stats = []
        dp_kni_stats_dict = {}

	# Get DP KNI Intf stats
        cmd_get_stats = ['edge-appctl', '-t', '/var/run/vmware/edge/dpd.ctl', 'kni/show']
        output = self.run_command_with_time_check(cmd_get_stats)

        dp_kni_stats = json.loads(output)

        for stat in dp_kni_stats:
            tmp_dict = {}
            tmp_dict = stat.get('stats')
            if tmp_dict:
                for key, value in tmp_dict.items():
                    stat["stats_" + key] = value

            del stat['stats']
            dp_kni_stats_dict[stat["ifuuid"]] = stat

        self._logger.debug("edge_dp_kni_intf stats({}): {}".format(sys.getsizeof(dp_kni_stats_dict), dp_kni_stats_dict))
        return dp_kni_stats_dict

    def get_dp_phy_intf_stats(self):
        dp_phy_stats = []
        dp_phy_stats_dict = {}

	# Get DP KNI Intf stats
        cmd_get_stats = ['edge-appctl', '-t', '/var/run/vmware/edge/dpd.ctl', 'physical_port/show', 'stats']
        output = self.run_command_with_time_check(cmd_get_stats)

        dp_phy_stats = json.loads(output)

        for stat in dp_phy_stats:
            dp_phy_stats_dict[stat["name"]] = stat

        self._logger.debug("edge_dp_phy_intf stats({}): {}".format(sys.getsizeof(dp_phy_stats_dict), dp_phy_stats_dict))
        return dp_phy_stats_dict

    def get_dp_phy_nic_throughput(self):
        try:
            with open(_EDGE_NIC_THROUGHPUT_FILE, "r") as f:
                fcntl.flock(f.fileno(), fcntl.LOCK_SH)
                context = json.load(f)
        except Exception as e:
            self._logger.error("Failed to read {}, error: {}"
                .format(_EDGE_NIC_THROUGHPUT_FILE, e))
            return None

        data = {}
        data["throughput_data"] = []
        nic_throughput_data = context.get("context", [])

        for nic_info in nic_throughput_data:
            if "edge_nic_name" not in nic_info or \
                "nic_throughput" not in nic_info:
                continue

            nic_name = nic_info.get("edge_nic_name")
            data["sample_time"] = context.get("sample_time")
            stat_name = "edge.phys." + nic_name + ".throughput_percent"
            data["throughput_data"].append({stat_name: nic_info.get("nic_throughput")})

        return None if not data else (self.wavefront_source, data)


    @timeout_cache(10)
    def usage(self, *args, **kwargs):
        dp_cpu_core_stats = {}
        dp_mem_pool_stats = {}
        dp_kni_intf_stats = {}
        dp_phy_intf_stats = {}
        dp_phy_nic_throughput_stats = {}

        try:
            dp_cpu_core_stats = self.get_dp_cpu_stats()
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get edge DP CPU core stats: {}, {}".format(e, trace))

        try:
            dp_mem_pool_stats = self.get_dp_mem_pool_stats()
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get edge DP Mem Pool stats: {}, {}".format(e, trace))

        try:
            dp_kni_intf_stats = self.get_dp_kni_intf_stats()
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get edge DP KNI Intf stats: {}, {}".format(e, trace))

        try:
            dp_phy_intf_stats = self.get_dp_phy_intf_stats()
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get edge DP Phy Intf stats: {}, {}".format(e, trace))

        try:
            dp_phy_nic_throughput_stats = self.get_dp_phy_nic_throughput()
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get edge DP Phy Intf stats: {}, {}".format(e, trace))

        return {'cpu_stats': dp_cpu_core_stats, \
                'mem_pool_stats': dp_mem_pool_stats, \
                'kni_intf_stats': dp_kni_intf_stats, \
                'phy_intf_stats': dp_phy_intf_stats, \
                'phy_inf_throughput_stats': dp_phy_nic_throughput_stats}
