# Copyright (C) 2020-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import sched
import threading
import time
import traceback

import sha.config
import sha.preconfig

from concurrent.futures import ThreadPoolExecutor

from sha.constants import METRICS_STR
from sha.core.whiteboard import DataWhiteboard
from sha.core.collector._collector import Collector

# priority of metric collectors
METRIC_COLLECTOR_PRIORITY_DEF = 1
METRIC_COLLECTOR_PRIORITY_LOW = 2

# index of item in metric
METRIC_INDEX_KEY = 0
METRIC_INDEX_FUTURE = 1
METRIC_INDEX_LAST_RUN_TIME = 2
METRIC_INDEX_LAST_DONE_TIME = 3
METRIC_INDEX_SEC_TO_NEXT = 4
METRIC_INDEX_PRIOEITY = 5
METRIC_INDEX_TYPE = 6

# metrics types
METRIC_TYPE_DEFAULT = 0
METRIC_TYPE_DEDICATED_THREAD = 1

# max count of SHA metrics.
SHA_METRIC_MAX_COUNT = 100

# stats
METRIC_COLLECT_TASK_SKIP = "task-skip"
METRIC_COLLECT_TASK_SUBMITTED = "task-submitted"
METRIC_COLLECT_TASK_DONE = "task-done"
METRIC_COLLECT_COLLECTED = "metric-collected"
METRIC_COLLECT_REPORTED = "metric-reported"
METRIC_COLLECT_FAIL_REPORT = "fail-report"
METRIC_COLLECT_FAIL_ACTION = "fail-action"
METRIC_COLLECT_FAIL_COLLECT = "fail-collect"


class MetricCollector(threading.Thread):

    def __init__(self, *args, **kwargs):
        super().__init__(name='nsx-sha-metric-collector', *args, **kwargs)
        self._is_running = False
        self._logger = sha.preconfig.get_sub_logger(__name__, s2comp = 'metric-collector')
        self._metric_lock = threading.Lock()
        self._metrics = [[metric_key, None, 0, 0, 0, METRIC_COLLECTOR_PRIORITY_DEF, \
                          METRIC_TYPE_DEDICATED_THREAD if metric_key in sha.config.DEDICATED_THREAD_METRICS \
                              else METRIC_TYPE_DEFAULT] \
                         for metric_key in sha.config.ENABLED_METRICS.keys()]
        self._scheduler = None
        self._executor = None
        self._stats = {METRICS_STR : {metric_key : {} for metric_key in sha.config.ENABLED_METRICS.keys()}}

    def __repr__(self):
        return ('MetricCollector_is_running:{}'.format(self._is_running))

    def _stats_inc(self, field, metric_key = None):
        if metric_key:
            stats = self._stats[METRICS_STR][metric_key]
        else:
            stats = self._stats
        stats[field] = stats.get(field, 0) + 1

    def _schedule_next_run(self, metric):
        # dedicated thread metric, not to use schedule
        if metric[METRIC_INDEX_TYPE] == METRIC_TYPE_DEDICATED_THREAD:
             return
        self._scheduler.enter(1, metric[METRIC_INDEX_PRIOEITY], self._schedule_action, argument=([metric]))

    def _schedule_action(self, metric):
        with self._metric_lock:
            self._schedule_action_with_lock(metric)

    def _schedule_action_with_lock(self, metric):
        """
            action of one schedule run
            Input: [metric_key, future, last_begin, last_complete, seconds to next run, priority, type]
        """
        if not self._is_running:
            return
        if metric[METRIC_INDEX_SEC_TO_NEXT]:
            metric[METRIC_INDEX_SEC_TO_NEXT] -= 1
            self._schedule_next_run(metric)
            return

        metric_usage, interval = Collector.get_metric(metric[METRIC_INDEX_KEY])
        if not metric_usage:
            # metric is not registered, schedule next check with low priority
            metric[METRIC_INDEX_SEC_TO_NEXT] = sha.config.GLOBAL_MINIMUM_CHECKING_INTERVAL - 1
            metric[METRIC_INDEX_PRIOEITY] = METRIC_COLLECTOR_PRIORITY_LOW
            self._schedule_next_run(metric)
            # self._logger.debug("No plugin need metric {}, skip it". format(metric[0]))
            return

        # check last run
        if metric[METRIC_INDEX_FUTURE]:
            if (metric[METRIC_INDEX_TYPE] == METRIC_TYPE_DEFAULT and \
                not metric[METRIC_INDEX_FUTURE].done()):
                # last run not complete yet, skip this run
                # schedule next check after interval seconds with low priority
                metric[METRIC_INDEX_SEC_TO_NEXT] = interval - 1
                metric[METRIC_INDEX_PRIOEITY] = METRIC_COLLECTOR_PRIORITY_LOW
                self._schedule_next_run(metric)
                self._logger.warning("Metric {} last execution {} not complete, running for {} secs"
                                     .format(metric[METRIC_INDEX_KEY],
                                             metric[METRIC_INDEX_FUTURE],
                                             time.monotonic() - metric[METRIC_INDEX_LAST_RUN_TIME]))
                self._stats_inc(METRIC_COLLECT_TASK_SKIP, metric[METRIC_INDEX_KEY])
                return
            # last run complete time
            if metric[METRIC_INDEX_TYPE] == METRIC_TYPE_DEFAULT:
                metric[METRIC_INDEX_LAST_DONE_TIME] = metric[METRIC_INDEX_FUTURE].result()
                self._logger.debug("Metric {} last execution {} complete: running for {} secs from {}"
                                   .format(metric[METRIC_INDEX_KEY],
                                           metric[METRIC_INDEX_FUTURE],
                                           metric[METRIC_INDEX_LAST_DONE_TIME] - metric[METRIC_INDEX_LAST_RUN_TIME],
                                           metric[METRIC_INDEX_LAST_RUN_TIME]))
                self._stats_inc(METRIC_COLLECT_TASK_DONE, metric[METRIC_INDEX_KEY])

        # submit task for default metrics
        metric[METRIC_INDEX_LAST_RUN_TIME] = time.monotonic()
        if metric[METRIC_INDEX_TYPE] == METRIC_TYPE_DEFAULT:
            # delete last done task
            if metric[METRIC_INDEX_FUTURE]:
                old_task = metric[METRIC_INDEX_FUTURE]
                metric[METRIC_INDEX_FUTURE] = None
                del old_task
            # submit the new task
            metric[METRIC_INDEX_FUTURE] = self._executor.submit(self._collect, metric[METRIC_INDEX_KEY], metric_usage)
            self._logger.debug("Metric {}: sumbit collect task {} at {}"
                               .format(metric[METRIC_INDEX_KEY], metric[METRIC_INDEX_FUTURE],
                                       metric[METRIC_INDEX_LAST_RUN_TIME]))
            self._stats_inc(METRIC_COLLECT_TASK_SUBMITTED, metric[METRIC_INDEX_KEY])
        else:
            #dedicated thread to collect metric, exeucte collect action directly.
            metric[METRIC_INDEX_LAST_DONE_TIME] = self._collect(metric[METRIC_INDEX_KEY], metric_usage)
            self._stats_inc(METRIC_COLLECT_TASK_DONE, metric[METRIC_INDEX_KEY])

        # schedule next check after interval seconds with default priority
        metric[METRIC_INDEX_SEC_TO_NEXT] = interval - 1
        metric[METRIC_INDEX_PRIOEITY] = METRIC_COLLECTOR_PRIORITY_DEF
        self._schedule_next_run(metric)

    def _collect(self, metric_key, metric_usage):
        self._logger.debug("Begin to collect metric: {}".format(metric_key))
        try:
            data = metric_usage()
            self._logger.debug("Collected metric data: {}".format(data))
            self._stats_inc(METRIC_COLLECT_COLLECTED, metric_key)
            if not DataWhiteboard.report_data(metric_key, data):
                self._logger.warning("Failed to report metric {} to whiteboard".format(metric_key))
                self._stats_inc(METRIC_COLLECT_FAIL_REPORT, metric_key)
            else:
                self._stats_inc(METRIC_COLLECT_REPORTED, metric_key)
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to collect metrics, Exception:{}, trace:{}".format(e, trace))
            self._stats_inc(METRIC_COLLECT_FAIL_COLLECT, metric_key)
        finally:
            return time.monotonic()

    def _dedicated_thread_collect(self, metric):
        self._logger.info('starting dedicated thread for metric {}'.format( metric[METRIC_INDEX_KEY]))
        while self._is_running:
            time.sleep(1)
            try:
                self._schedule_action(metric)
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("Failed to execut schedule action, Exception:{}, trace:{}".format(e, trace))
                self._stats_inc(METRIC_COLLECT_FAIL_ACTION, metric[METRIC_INDEX_KEY])

        Collector.release_metric_resource(metric[METRIC_INDEX_KEY])

        self._logger.warning('stopped dedicated thread for metric {}'.format( metric[METRIC_INDEX_KEY]))

    def run(self):
        self._logger.info('starting metric collector')
        self._is_running = True
        # Use SHA_METRIC_MAX_COUNT to make sure each metric can be executed in time.
        max_workers = min(SHA_METRIC_MAX_COUNT, len(sha.config.ENABLED_METRICS))
        self._logger.info('Thread pool max worker count is: {}'.format(max_workers))
        self._executor = ThreadPoolExecutor(max_workers=max_workers)
        self._scheduler = sched.scheduler(time.monotonic, time.sleep)
        with self._metric_lock:
            for metric in self._metrics:
                if metric[METRIC_INDEX_TYPE] == METRIC_TYPE_DEFAULT:
                    self._schedule_next_run(metric)
                else:
                    metric[METRIC_INDEX_FUTURE] = threading.Thread(target=self._dedicated_thread_collect,
                                                                   args=([metric]),daemon=True)
                    metric[METRIC_INDEX_FUTURE].start()
                    self._logger.debug("Metric {}: dedicated thread to collect {}"
                                       .format(metric[METRIC_INDEX_KEY], metric[METRIC_INDEX_FUTURE]))

        self._scheduler.run()

        self._logger.warning("stopped metric collector")
        # Dump metric status and stats
        with self._metric_lock:
            for metric in self._metrics:
                self._logger.info("metric: {}, stats: {}".
                                  format(metric, self._stats[METRICS_STR][metric[METRIC_INDEX_KEY]]))

    def stop(self):
        self._logger.warning("going to stop metric collector")
        self._is_running = False
        with self._metric_lock:
            for metric in self._metrics:
                if metric[METRIC_INDEX_TYPE] == METRIC_TYPE_DEFAULT and \
                   metric[METRIC_INDEX_FUTURE]:
                    self._logger.warning("going to cancel metric {}".format(metric))
                    metric[METRIC_INDEX_FUTURE].cancel()
                elif metric[METRIC_INDEX_TYPE] == METRIC_TYPE_DEDICATED_THREAD and \
                     metric[METRIC_INDEX_FUTURE]:
                    self._logger.warning("going to stop metric dedicated thread: {}".format(metric))
                    # dedicated thread will check self._is_running and exit

    def wait_stop(self):
        #wait for dedicated threads done
        with self._metric_lock:
            for metric in self._metrics:
                if metric[METRIC_INDEX_TYPE] == METRIC_TYPE_DEDICATED_THREAD and \
                   metric[METRIC_INDEX_FUTURE]:
                    metric[METRIC_INDEX_FUTURE].join()
        # wait for pending futures done
        if self._executor:
            self._executor.shutdown()
        self._logger.warning("stopped all metric task executors")
        #wait for metric collector own thread done.
        self.join()
