# Copyright (C) 2020-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
from threading import Thread, Lock

import os
import signal
import subprocess
import sys
import time
import traceback

import sha.preconfig
from sha.constants import PLUGIN, PLUGIN_ID, KEY, VALUE, TIMESTAMP, DATA, EVENT_QUEUE, \
                          METRICS_STR, PREFIX_METRICS_STR, CHILDREN

from concurrent.futures import ThreadPoolExecutor
from queue import Queue, Empty, Full
from copy import deepcopy

# Data Report:
# {
#    'key': key of the data,
#    'value': the data value,
#    'timestamp': timestamp of the data collected
# }
#
# Data Request/Data Registration prop:
# {
#   # UUid of the plugin
#   'plugin_id': plugin id,
#   # list of the requested data
#   # each data is located based on key of the data(name of the metric)
#   'data': {
#       'metrics':[
#           key1,
#           key2
#       ],
#       'prefix_metrics':[
#           prefix_key1,
#       ],
#   },
#   # Event queue to notify plugin
#   'event_queue': queue
# }
#
# Data notify event to notifier
# {
#   'plugin': {
#        plugin_id1: event_queue1,
#        plugin_id2: event_queue2,
#   },
#   # list of the requested data
#   # each data is located based on the namespace of the reporter and key of the data
#   'data': {
#       key1 : {
#           'value': value1,
#           'timestamp': timestamp1
#       },
#       key2 : {
#           'value': value2,
#           'timestamp': timestamp2
#       }
#   ],
# }
#
# Data notify event to plugin:
# {
#   'plugin_id': plugin_id,
#   # list of the requested data
#   # each data is located based on the namespace of the reporter and key of the data
#   'data': [
#       key1 : {
#           'value': value1,
#           'timestamp': timestamp1
#       },
#       key2 ： {
#           'value': value2,
#           'timestamp': timestamp2
#       }
#   ],
# }

# The queues to receive data report from metric and data request from plugin
DATA_REPORT_QUEUE_SIZE = 100
DATA_REQUEST_QUEUE_SIZE = 100
# The internal queue to send the prepared data to the notifer
DATA_NOTIFY_QUEUE_SIZE = 200

# timeout seconds when reporting/requesting data
DATA_WHITEBOARD_REPORT_TIMEOUT = 1
DATA_WHITEBOARD_REQUEST_TIMEOUT = 1
DATA_WHITEBOARD_DATA_TIMEOUT = 2

# Data whiteboard stats counter
#report listener
DATA_WHITEBOARD_STATS_REPORT_RCVD   = "report-received"
DATA_WHITEBOARD_STATS_REPORT_OK     = "report-stored"
DATA_WHITEBOARD_STATS_REPORT_FAIL   = "report-fail"
DATA_WHITEBOARD_STATS_REPORT_NOTIFY = "report-notify-sent"

#request listener
DATA_WHITEBOARD_STATS_REQUEST_RCVD    = "request-received"
DATA_WHITEBOARD_STATS_REQUEST_INVALID = "request-invalid"
DATA_WHITEBOARD_STATS_REQUEST_FAIL    = "request-fail"
DATA_WHITEBOARD_STATS_REQUEST_NOTIFY  = "request-notify-sent"

# notifier
DATA_WHITEBOARD_STATS_NOTIFY_RCVD = "notify-received"
DATA_WHITEBOARD_STATS_NOTIFY_FAIL = "notify-fail-other"
DATA_WHITEBOARD_STATS_NOTIFY_FULL = "notify-fail-queue-full"
DATA_WHITEBOARD_STATS_NOTIFY_SENT = "notify-sent"

# others
DATA_WHITEBOARD_STATS_REPORT_DATA       = "report-data"
DATA_WHITEBOARD_STATS_REPORT_DATA_FULL  = "report-data-fail-queue-full"
DATA_WHITEBOARD_STATS_REPORT_DATA_FAIL  = "report-data-fail-other"
DATA_WHITEBOARD_STATS_REQUEST_DATA      = "request-data"
DATA_WHITEBOARD_STATS_REQUEST_DATA_FULL = "request-data-fail-queue-full"
DATA_WHITEBOARD_STATS_REQUEST_DATA_FAIL = "request-data-fail-other"
DATA_WHITEBOARD_STATS_GET_DATA          = "get-data"
DATA_WHITEBOARD_STATS_GET_DATA_EMPTY    = "get-data-fail-queue-empty"
DATA_WHITEBOARD_STATS_GET_DATA_FAIL     = "get-data-fail-other"

# DUMP DATA
import tempfile
DUMP_FILE_PATH = os.path.join(tempfile.gettempdir(), "nsx-sha-data")

class DataWhiteboardTree:
    """
    Dictionary tree of sub-keys splited by "."
    Each node has the "children" field. Its value of the leaf node is {}.
    Each node may have its own "data" field.

    E.g:
    {
        'children': {
            'tsdb' : {
                'data': {
                     'value': value2
                     'timestamp': timestamp2,
                },
                'children': {
                    'edge': {
                        'data': {
                            'value': value3,
                            'timestamp': timestamp3,
                        },
                        'children': {
                            'status': {
                                'data' :{
                                    'value': value4,
                                    'timestamp': timestamp4,
                                },
                                'children': {}
                            }
                        }
                    }
                },
            },
            'nsx' : {
                'children': {
                    'uplink': {
                        'children': {
                            'status': {
                                'data' : {
                                    'value': value1,
                                    'timestamp': timestamp1,
                                }
                                'children': {}
                            }
                        }
                    }
                }
            }
        }
    }
    """
    def __init__(self):
        self._logger = sha.preconfig.get_sub_logger(__name__, s2comp = 'data-whiteboard-tree')
        self._tree = {CHILDREN:{}}


    # Dictionary tree traversal
    def _get_all_data(self, prefix_path, cur_node):
        data = {}
        if DATA in cur_node:
             data['.'.join(prefix_path)] = cur_node[DATA]
        # Traversal children
        for key in cur_node[CHILDREN]:
            data.update(self._get_all_data(prefix_path + [key], cur_node[CHILDREN][key]))

        self._logger.debug("Get data from all nodes under path {} : {}".format(prefix_path, data))
        return data

    # Get data from all nodes start with the prefix_key
    def get_data_start_with(self, prefix_key):
        prefix_path = prefix_key.split('.')
        # For the key "XXX.", we will get path ['XXX', '']. Need ignore the last empty sub_key
        if prefix_path[-1] == '':
            del prefix_path[-1]
        cur_node = self._tree
        for key in prefix_path:
            try:
                cur_node = cur_node[CHILDREN][key]
            except:
                self._logger.debug("No metric match prefix {}, return directly".format(prefix_key))
                return {}
        return self._get_all_data(prefix_path, cur_node)

    # Get data from all nodes perfix match the key
    def get_data_on_the_path(self, key):
        path = key.split('.')
        prefix_path = []
        cur_node = self._tree
        data = {}
        for key in path:
            if DATA in cur_node:
                data['.'.join(prefix_path)]= cur_node[DATA]
            try:
                cur_node = cur_node[CHILDREN][key]
                prefix_path.append(key)
            except:
                # hit leaf node, return
                return data

    # Get data from one specified node
    def get_data(self, key):
        path = key.split('.')
        cur_node = self._tree
        for key in path:
            try:
                cur_node = cur_node[CHILDREN][key]
            except:
                # hit leaf node, return
                self._logger.debug("No metric match {}".format(key))
                return None
        return cur_node.get(DATA, None)

    # Set data into the dictionary tree
    def set_data(self, key, data):
        path = key.split('.')
        # For the key "XXX.", we will get path ['XXX', '']. Need ignore the last empty sub_key
        if path[-1] == '':
            del path[-1]
        self._logger.debug("Set data on key {} with path {}".format(key, path))
        cur_node = self._tree
        for key in path:
            try:
                cur_node = cur_node[CHILDREN][key]
            except:
                child_node = {CHILDREN: {}}
                cur_node[CHILDREN][key] = child_node
                cur_node = child_node
        cur_node[DATA] = data
        self._logger.debug("Set data to the specified node '{} : {}'".format(key, data))
        self._logger.debug('data tree: {}'.format(self._tree))

    # Clear data from the dictionary tree
    def clear_data(self, key):
        path = key.split('.')
        cur_node = self._tree
        node_list = []
        for key in path:
            try:
                node_list.append(cur_node)
                cur_node = cur_node[CHILDREN][key]
            except:
                self._logger.warning("Failed to find the node of key {} ".format(key))
                return
        del cur_node[DATA]
        self._logger.debug("clear data on the specified node {}".format(key))
        if cur_node[CHILDREN]:
            return

        # The leaf node is cleared, need check the whole path to clear
        i = len(node_list) - 1
        while i >= 0:
            cur_node = node_list[i]
            key = path[i]
            del cur_node[CHILDREN][key]
            self._logger.debug("Clear the empty node {}".format(path[0:i].join('.')))
            if cur_node[CHILDREN] or cur_node.get(DATA, None):
                return

    @property
    def data(self):
        return self._tree

class DataWhiteboardStore:
    """ Data Store of white board
        _data_base: key, value paires
        E.g:
        {
            'nsx.uplink.uplink-status': {
                'value': value1,
                'timestamp': timestamp1,
            },
            'tsdb': {
                'value': value2,
                'timestamp': timestamp2,
            },
            'tsdb.edge': {
                'value': value3,
                'timestamp': timestamp3,
            },
            'tsdb.edge.status': {
                'value': value4,
                'timestamp': timestamp4,
            }
        }

        _data_base_tree: tree for prefix key query

    """
    def __init__(self):
        self._logger = sha.preconfig.get_sub_logger(__name__, s2comp = 'data-whiteboard-store')
        self._data_base = {}
        self._data_base_tree = DataWhiteboardTree()
        self._data_base_lock = Lock()

    def set(self, key, value, timestamp=None):
        with self._data_base_lock:
            self._logger.debug('Existing metric data {}: {}'
                               .format(key, self._data_base.get(key, None)))
            # Update value, timestamp of the data
            self._data_base[key] = {VALUE: value, TIMESTAMP: timestamp}
            self._logger.debug('Updated metric data {} : {}'
                               .format(key, self._data_base[key]))
            self._data_base_tree.set_data(key, self._data_base[key])
            return self._data_base[key]

    def get(self, key):
        with self._data_base_lock:
            self._logger.debug('Existing full match metric data {}: {}'
                               .format(key, self._data_base.get(key, None)))
            return self._data_base.get(key, None)

    def get_prefix(self, prefix_key):
        with self._data_base_lock:
            data = self._data_base_tree.get_data_start_with(prefix_key)
            self._logger.debug('Existing prefix match metric data {}: {}'
                               .format(prefix_key, data))
            return data

    def dump(self, file_path):
        # Dump to pickle file
        import pickle
        try:
            self._logger.info("Try to dump data to pickle file: {}".format(file_path+".pkl"))
            with open(file_path+'.pkl', 'wb') as outfile:
                with self._data_base_lock:
                    pickle.dump([self._data_base, self._data_base_tree.data], outfile)
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed when dumping data to pickle file, Exception:{}, trace:{}".format(e, trace))

            # try to dump the text file instead:
            self._logger.warning("Try to dump data to text file: {}".format(file_path+".txt"))
            try:
                with open(file_path+".txt", 'w') as textfile:
                    with self._data_base_lock:
                        textfile.write("whiteboard data store: {}\n".format([self._data_base, self._data_base_tree.data], compact=True))
                        if self._data_base.get('nsx.daemon.daemon-health', None):
                            textfile.write("'nsx.daemon.daemon-health': {}".format(self._data_base['nsx.daemon.daemon-health']['value']))
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("Failed when dumping data to text file, Exception:{}, trace:{}".format(e, trace))


class DataRegister:
    """ Data Register maps of white board
        E.g:
        1. plugin_metric_map: all registered plugins
        {
              plugin_id1: {
                  'data': {
                       'metrics': [key1, key2],
                  },
                  'event_queue': event_queue1
              },
              plugin_id2: {
                  'data': {
                      'metrics': [key2],
                      'prefix_metrics': [prefix_key1],
                  },
                  'event_queue': event_queue2
              },
        }
        2. metric_plugin_map: full match metrics
        {
           key1: {
              plugin_id1: event_queue1,
           },
           key2: {
              plugin_id1: event_queue1,
              plugin_id2: event_queue2,
           },
        }
        3. prefix_metric_plugin_map: tree for prefix match metrics
    """
    def __init__(self):
        self._logger = sha.preconfig.get_sub_logger(__name__, s2comp = 'data-whiteboard-registration')
        self._plugin_metric_map = {}
        self._metric_plugin_map = {}
        self._prefix_metric_plugin_map = DataWhiteboardTree()
        self._plugin_metric_lock = Lock()

    def register(self, plugin_id, metrics, prefix_metrics, event_queue):
        # Validate metrics
        if not metrics and not prefix_metrics:
            self._logger.warning('Invalid metrics and prefix_metrics.')
            raise Exception('Invalid metrics')
        if not event_queue:
            self._logger.warning('Invalid event_queue')
            raise Exception('Invalid event_queue')

        metrics =  metrics if metrics else []
        prefix_metrics = prefix_metrics if prefix_metrics else []
        with self._plugin_metric_lock:
            legacy_metrics = self._plugin_metric_map.get(plugin_id, None)
            if legacy_metrics:
                self._logger.warning('Plugin {} has registered before: {}'.format(plugin_id, legacy_metrics))
                raise Exception('Plugin has registed'.format(plugin_id))
            # plugin-metric mapping
            prop  = {
                DATA: {
                    METRICS_STR: metrics,
                    PREFIX_METRICS_STR: prefix_metrics,
                },
                EVENT_QUEUE: event_queue
            }
            self._plugin_metric_map[plugin_id] = prop
            # metric-plugin mapping
            for metric in metrics:
                item = self._metric_plugin_map.get(metric, None)
                if item:
                    item[plugin_id] = event_queue
                else:
                    self._metric_plugin_map[metric] = {plugin_id: event_queue}

            for prefix_metric in prefix_metrics:
                item = self._prefix_metric_plugin_map.get_data(prefix_metric)
                self._logger.debug("Get existing plugins from prefix_metric_plugin_map {} : {}".format(prefix_metric, item))
                if item:
                    item[plugin_id] = event_queue
                else:
                    self._prefix_metric_plugin_map.set_data(prefix_metric, {plugin_id: event_queue})

        self._logger.info('Plugin {} registered successfully with metrics: {} and prefix_metrics: {}'
                          .format(plugin_id, metrics, prefix_metrics))
        self._logger.debug('Plugin-metric map: {}'.format(self._plugin_metric_map))
        self._logger.debug('Metric-plugin map: {}'.format(self._metric_plugin_map))


    def unregister(self, plugin_id):
        with self._plugin_metric_lock:
            prop = self._plugin_metric_map.pop(plugin_id, None)
            if not prop:
                self._logger.info('Plugin {} has not registered.'.format(plugin_id))
                return
            for metric in prop[DATA][METRICS_STR]:
                try:
                    del self._metric_plugin_map[metric][plugin_id]
                    if not self._metric_plugin_map[metric]:
                        del self._metric_plugin_map[metric]
                except Exception as e:
                    trace = traceback.format_exc()
                    self._logger.warning("Failed in DataWhiteboard, Exception:{}, trace:{}".format(e, trace))
            for prefix_metric in prop[DATA][PREFIX_METRICS_STR]:
                try:
                    item = self._prefix_metric_plugin_map.get_data(prefix_metric)
                    self._logger.debug("Get existing plugins from prefix_metric_plugin_map {} : {}".format(prefix_metric, item))
                    del item[plugin_id]
                    if not item:
                        self._prefix_metric_plugin_map.clear_data(prefix_metric)
                except Exception as e:
                    trace = traceback.format_exc()
                    self._logger.warning("Failed in DataWhiteboard, Exception:{}, trace:{}".format(e, trace))

        self._logger.info('Plugin {} unregistered with {}'.format(plugin_id, prop[DATA]))

        self._logger.info('Plugin {} unregistered with {}'.format(plugin_id, prop[DATA]))
        self._logger.debug('Plugin-metric map: {}'.format(self._plugin_metric_map))
        self._logger.debug('Metric-plugin map: {}'.format(self._metric_plugin_map))

    def get_plugins(self, key):
        with self._plugin_metric_lock:
            # get plugins registered full match
            full_plugins = self._metric_plugin_map.get(key, None)
            self._logger.debug('Registered plugins for metric {}: {}'
                               .format(key, full_plugins))
            #get plugins registered prefix match
            prefix_plugins = self._prefix_metric_plugin_map.get_data_on_the_path(key)
            self._logger.debug('Registered plugins prefix match metric {}: {}'
                               .format(key, prefix_plugins))
            all_plugins = full_plugins.copy() if full_plugins else {}
            for plugins in prefix_plugins.values():
                all_plugins.update(plugins)

            self._logger.debug('All registered plugins for metric {}: {}'
                               .format(key, all_plugins))
            return all_plugins


class DataWhiteboard:
    _logger = sha.preconfig.get_sub_logger(__name__, s2comp = 'data-whiteboard')
    _data_report_queue =  Queue(DATA_REPORT_QUEUE_SIZE)
    _data_request_queue =  Queue(DATA_REPORT_QUEUE_SIZE)
    _data_notify_queue =  Queue(DATA_REPORT_QUEUE_SIZE)
    _data_store = DataWhiteboardStore()
    _data_register= DataRegister()
    _report_listener = None
    _request_listener = None
    _data_notifier = None
    _is_running = False
    _stats = {
    }

    @classmethod
    def _stats_inc(cls, field):
        cls._stats[field] = cls._stats.get(field, 0) + 1

    @classmethod
    def start(cls):
        # Start the listeners and notifier
        cls._is_running = True
        cls._report_listener = Thread(target=cls._report_listen, daemon=True)
        cls._request_listener = Thread(target=cls._request_listen, daemon=True)
        cls._data_notifier = Thread(target=cls._data_notify, daemon=True)
        cls._report_listener.start()
        cls._request_listener.start()
        cls._data_notifier.start()


    @classmethod
    def stop(cls):
        cls._logger.warning("going to stop data whiteboard")
        cls._is_running = False

    @classmethod
    def wait_stop(cls):
        cls._report_listener.join()
        cls._request_listener.join()
        cls._data_notifier.join()

    @classmethod
    def _report_listen(cls):
        cls._logger.info('starting report listener')
        while cls._is_running:
            try:
                report = cls._data_report_queue.get(timeout = 1)
                cls._logger.debug('Receive data report: {}'.format(report))
                cls._stats_inc(DATA_WHITEBOARD_STATS_REPORT_RCVD)

                # Save data in data store
                data = cls._data_store.set(report[KEY], report[VALUE], report.get(TIMESTAMP, None))
                cls._stats_inc(DATA_WHITEBOARD_STATS_REPORT_OK)

                # Send data to notifier if it is a monitor metric data
                plugins = cls._data_register.get_plugins(report[KEY])
                if plugins:
                    event = {
                        PLUGIN: plugins,
                        DATA: {report[KEY]: data}
                    }
                    cls._data_notify_queue.put(event)
                    cls._logger.debug('Send event to notifier: {}'.format(event))
                    cls._stats_inc(DATA_WHITEBOARD_STATS_REPORT_NOTIFY)

            except Empty:
                continue
            except Exception as e:
                trace = traceback.format_exc()
                cls._logger.warning("Failed in data report listener, Exception:{}, trace:{}".format(e, trace))
                cls._stats_inc(DATA_WHITEBOARD_STATS_REPORT_FAIL)

        # Dump data to file
        cls._logger.info('Dump data store to file: {}'.format(DUMP_FILE_PATH))
        cls._data_store.dump(DUMP_FILE_PATH)
        cls._logger.warning("stopped whiteboard data report listener")


    @classmethod
    def _request_listen(cls):
        cls._logger.info('starting request listener')
        while cls._is_running:
            try:
                request = cls._data_request_queue.get(timeout = 1)
                cls._logger.debug('Receive request: {}'.format(request))
                cls._stats_inc(DATA_WHITEBOARD_STATS_REQUEST_RCVD)

                if not request.get(EVENT_QUEUE, None):
                    cls._logger.warning('Invalid event_queue in request {}'.format(request))
                    cls._stats_inc(DATA_WHITEBOARD_STATS_REQUEST_INVALID)
                    continue

                # get data from data store
                data_dict = {}
                for key in request.get(DATA, {}).get(METRICS_STR, []):
                    data_dict[key] = cls._data_store.get(key)
                cls._logger.debug('Get full match metrics: {}'.format(data_dict))
                for prefix_key in request.get(DATA, {}).get(PREFIX_METRICS_STR, []):
                    data_dict.update(cls._data_store.get_prefix(prefix_key))
                cls._logger.debug('Get data_dict: {}'.format(data_dict))

                # Send data to notifer
                event = {
                    PLUGIN: {request[PLUGIN_ID] : request[EVENT_QUEUE]},
                    DATA: data_dict
                }
                cls._data_notify_queue.put(event)
                cls._logger.debug('Send event to notifier: {}'.format(event))
                cls._stats_inc(DATA_WHITEBOARD_STATS_REQUEST_NOTIFY)

            except Empty:
                continue
            except Exception as e:
                trace = traceback.format_exc()
                cls._logger.warning("Failed in data request listner, Exception:{}, trace:{}".format(e, trace))
                cls._stats_inc(DATA_WHITEBOARD_STATS_REQUEST_FAIL)

        cls._logger.warning("stopped whiteboard data request listener")

    @classmethod
    def _data_notify(cls):
        cls._logger.info('starting data notifier')
        while cls._is_running:
            try:
                event = cls._data_notify_queue.get(timeout = 1)
                cls._logger.debug('Get notify event: {}'.format(event))
                cls._stats_inc(DATA_WHITEBOARD_STATS_NOTIFY_RCVD)

                # Send event to notify each plugin
                for plugin_id, event_queue in event[PLUGIN].items():
                    try:
                        # make a copy of data for each plugin
                        plugin_event = {
                                PLUGIN_ID: plugin_id,
                                DATA: deepcopy(event[DATA])
                        }
                        event_queue.put(plugin_event, timeout=0.1)
                        cls._logger.debug('Send event to plugin {}: {}'.format(plugin_id, plugin_event))
                        cls._stats_inc(DATA_WHITEBOARD_STATS_NOTIFY_SENT)
                    except Full:
                        cls._logger.warning("Failed to notify plugin {} due to queue is full".format(plugin_id))
                        cls._stats_inc(DATA_WHITEBOARD_STATS_NOTIFY_FULL)
                    except Exception as e:
                        trace = traceback.format_exc()
                        cls._logger.warning("Failed to notify plugin, Exception:{}, trace:{}".format(e, trace))
                        cls._stats_inc(DATA_WHITEBOARD_STATS_NOTIFY_FAIL)

            except Empty:
                continue
            except Exception as e:
                trace = traceback.format_exc()
                cls._logger.warning("Failed in data notifier, Exception:{}, trace:{}".format(e, trace))
                cls._stats_inc(DATA_WHITEBOARD_STATS_NOTIFY_FAIL)

        cls._logger.warning("stopped whiteboard data notifier")

    @classmethod
    def register(cls, plugin_id, metrics, event_queue, prefix_metrics=[]):
        cls._data_register.register(plugin_id, metrics, prefix_metrics, event_queue)

    @classmethod
    def unregister(cls, plugin_id):
        cls._data_register.unregister(plugin_id)

    @classmethod
    def get_plugin(cls, key):
        return cls._data_register.get_plugins(key)

    @classmethod
    def report_data(cls, key, value, timestamp = None):
        if not timestamp:
            timestamp = time.time()
        try:
            report = {KEY: key, VALUE: deepcopy(value), TIMESTAMP: timestamp}
            cls._data_report_queue.put(report,
                                       timeout=DATA_WHITEBOARD_REPORT_TIMEOUT)
            cls._logger.debug('Report data to whiteboard: {}'.format(report))
            cls._stats_inc(DATA_WHITEBOARD_STATS_REPORT_DATA)
        except Full:
            cls._logger.warning("Failed to report data {} due to timeout".format(key))
            cls._stats_inc(DATA_WHITEBOARD_STATS_REPORT_DATA_FULL)
            return False
        except Exception as e:
            trace = traceback.format_exc()
            cls._logger.warning("Failed to report data {}, Exception:{}, trace:{}".format(key, e, trace))
            cls._stats_inc(DATA_WHITEBOARD_STATS_REPORT_DATA_FAIL)
            raise e
        return True

    @classmethod
    def request_data(cls, plugin_id, metrics, event_queue, prefix_metrics=[]):
        try:
            request = {
                    PLUGIN_ID: plugin_id,
                    DATA: {
                        METRICS_STR: metrics,
                        PREFIX_METRICS_STR: prefix_metrics,
                    },
                    EVENT_QUEUE: event_queue
                }
            cls._data_request_queue.put(request,
                timeout = DATA_WHITEBOARD_REQUEST_TIMEOUT)
            cls._logger.debug('Send request to whiteboard: {}'.format(request))
            cls._stats_inc(DATA_WHITEBOARD_STATS_REQUEST_DATA)
        except Full:
            cls._logger.warning("Failed to request metric for plugin {} due to timeout".format(plugin_id))
            cls._stats_inc(DATA_WHITEBOARD_STATS_REQUEST_DATA_FULL)
            return False
        except Exception as e:
            trace = traceback.format_exc()
            cls._logger.warning("Failed to request data for plugin {}, Exception:{}, trace:{}".format(plugin_id, e, trace))
            cls._stats_inc(DATA_WHITEBOARD_STATS_REQUEST_DATA_FAIL)
            return False
        return True

    @classmethod
    def get_data(cls, event_queue, timeout=None):
        try:
            data_event = event_queue.get(timeout=timeout if timeout else DATA_WHITEBOARD_DATA_TIMEOUT)
            cls._logger.debug('Rcvd event from whiteboard: {}'.format(data_event))
            cls._stats_inc(DATA_WHITEBOARD_STATS_GET_DATA)
            return data_event[DATA]
        except Empty:
            cls._stats_inc(DATA_WHITEBOARD_STATS_GET_DATA_EMPTY)
            raise Empty
        except Exception as e:
            trace = traceback.format_exc()
            cls._logger.warning("Failed to get metrics, Exception:{}, trace:{}".format(e, trace))
            cls._stats_inc(DATA_WHITEBOARD_STATS_GET_DATA_FAIL)
            raise e

    @classmethod
    def get_metrics(cls, plugin_id, metrics, timeout=None, prefix_metrics=[]):
        event_queue = Queue(1)
        if not cls.request_data(plugin_id, metrics, event_queue, prefix_metrics=prefix_metrics):
            return None
        try:
            return cls.get_data(event_queue, timeout)
        except Empty:
            cls._logger.warning("Failed to request metrics for plugin {} due to timeout".format(plugin_id))
            return None
        except Exception as e:
            trace = traceback.format_exc()
            cls._logger.warning("Failed to get metrics, Exception:{}, trace:{}".format(e, trace))
            return None

    @classmethod
    def dump(cls, file_path):
        cls._logger.info("Dump data to {}".format(file_path))
        cls._data_store.dump(file_path)
